from typing import Union

import cv2

from .io import iterate_frames
from .processing import find_chessboard_corners, orient_board, locate_pieces, preprocess, color_postprocess, board_postprocess
from .models import load_models
from .game import StartChessGame


def run_video_to_pgn(
    video: Union[str, int],
    board_model: str,
    piece_model: str,
    out: str = "out.pgn",
    show: bool = False,
    device: str = "CPU",
) -> None:
    """
    Loads in models, iterates through frames while inferencing chess game state, and writes to PGN.
    """
    models, board_input_shapes, piece_input_shapes, board_precision, piece_precision = load_models(board_model, piece_model, device=device)
    game = StartChessGame(board_delay=40)

    for index, frame in iterate_frames(video, 640):
        piece_name_processed_frame, piece_pad = preprocess(frame, piece_input_shapes, piece_precision)
        piece_outputs = models.predict_color(piece_name_processed_frame)
        piece_results = color_postprocess(frame, piece_outputs, piece_pad, piece_input_shapes)
        if index == 0:
            board_processed_frame, board_pad = preprocess(frame, board_input_shapes, board_precision)
            board_outputs = models.predict_board(board_processed_frame)
            corner_results = board_postprocess(frame, board_outputs, board_pad, board_input_shapes)
            conversion_matrix, corners = find_chessboard_corners(corner_results)
            rotation = orient_board(piece_results, conversion_matrix)
        else:
            raw_map = locate_pieces(piece_results, conversion_matrix, rotation)
            game.update_board_stack(raw_map)
            game.update_move_stack()
            if game.board_has_changed():
                mismatched_squares = game.find_mismatches()
                game.generate_moves_from_mismatches(mismatched_squares)
                game.validate_moves_and_push()
                game.find_and_validate_takebacks_and_push()
                out_of_turn_moves = game.find_out_of_turn_moves()
                potentially_skipped_moves = game.find_potentially_skipped_moves(out_of_turn_moves)
                game.validate_potentially_skipped_moves_and_push(potentially_skipped_moves)

        if not show:
            continue
        for piece in piece_results:
            color = (0, 255, 0) if piece[4] == 0 else (0, 0, 255)
            cv2.circle(frame, (piece[0], piece[1]), 1, color, 5)
        if corners:
            for corner in corners:
                cv2.circle(frame, corner, 1, (255, 0, 0), 5)
        cv2.imshow("video", frame)
        if cv2.waitKey(1) & 0xff == ord('q'):
            break

    game.write_pgn(out)



from pathlib import Path
from typing import Optional

import typer

from .io import default_pgn_out_path
from .assets import CACHE_DIR, DEFAULT_MODELS, download_models, find_model, valid_model
from .pipeline import run_video_to_pgn


app = typer.Typer(no_args_is_help=True, add_completion=False, help="KnightVision CLI")
models_app = typer.Typer(help="Manage KnightVision models")
app.add_typer(models_app, name="models")


@models_app.command(help="Download model(s) from latest GitHub release")
def download(
    which: str = typer.Argument(..., help="'board', 'pieces', or 'all'"),
    dest: Path = typer.Option(CACHE_DIR / "models", help="Destination directory"),
):
    """
    Download model assets from the latest GitHub release into a local directory.
    """
    valid = list(DEFAULT_MODELS) + ["all"]
    if which not in valid:
        raise typer.BadParameter(f"Must be one of {valid}")
    names = list(DEFAULT_MODELS) if which == "all" else [which]
    msg = f"Download {', '.join(names)} to {dest}?"
    typer.confirm(msg, abort=True)
    for n in names:
        filename = DEFAULT_MODELS[n]
        download_models(filename, dest / filename)


@models_app.command(help="Search for models")
def locate():
    """
    Shows where each model would be resolved from by default.
    """
    for n in DEFAULT_MODELS:
        p = find_model(n)
        status = str(p) if p else "(not found)"
        typer.echo(f"{n} model: {status}")


@app.command(help="Process a video into PGN")
def run(
    video: Path = typer.Option(..., "--video", exists=True, readable=True, help="Path to input video"),
    board_model: Optional[Path] = typer.Option(None, "--board-model", help="Path to the board model"),
    piece_model: Optional[Path] = typer.Option(None, "--piece-model", help="Path to the piece model"),
    out: Optional[Path] = typer.Option(None, "--out", help="Path for output PGN"),
    show: bool = typer.Option(False, "--show/--no-show", help="Display windows"),
):
    """
    Runs video inferencing and outputs PGN.
    """
    out_path = out if out else default_pgn_out_path(video)
    out_path.parent.mkdir(parents=True, exist_ok=True)

    board_path = board_model if valid_model(model=board_model) else find_model('board')
    pieces_path = piece_model if valid_model(model=piece_model) else find_model('pieces')
    typer.echo(f"Using models:  {board_path}, {pieces_path}")

    run_video_to_pgn(video=str(video),
                     out=str(out_path),
                     board_model=str(board_path),
                     piece_model=str(pieces_path),
                     show=show)
    typer.echo(f"PGN saved at {out_path}")

if __name__ == "__main__":
    app()



from dataclasses import dataclass
from pathlib import Path
from typing import Sequence

import numpy as np
import onnxruntime as ort


def _pick_device(device: str = "CPU") -> list[str]:
    """
    Attempts to run on GPU if selected then falls back to CPU inferencing
    """
    if device == "GPU":
        if ort.get_device() == "GPU":
            ort.preload_dlls()
            return ["CUDAExecutionProvider", "CPUExecutionProvider"]
        else:
            print("GPU setup failed, falling back to CPU")
            return ["CPUExecutionProvider"]
    else:
        return ["CPUExecutionProvider"]


def _must_exist(path: Path) -> Path:
    """
    Checks existence of model path.
    """
    if not path.exists():
        raise FileNotFoundError(f"Model file not found: {path}\n")
    return path


@dataclass
class Models:
    """
    Thin container around onnx models.
    """
    board: ort.InferenceSession
    color: ort.InferenceSession
    board_inputs: Sequence[ort.NodeArg]
    color_inputs: Sequence[ort.NodeArg]

    def predict_board(self, img_data: np.ndarray):
        """
        Runs the board model on a frame.
        """
        return self.board.run(None, {self.board_inputs[0].name: img_data})

    def predict_color(self, img_data: np.ndarray):
        """
        Runs the piece color classifier on a frame.
        """
        return self.color.run(None, {self.color_inputs[0].name: img_data})


def load_models(board_model: str, color_model: str, device: str = "CPU") -> tuple[Models, tuple[int, int], tuple[int, int], bool, bool]:
    """
    Load both models once.
    """
    providers=_pick_device(device)

    board_path = Path(board_model)
    color_path = Path(color_model)
    board = ort.InferenceSession(str(_must_exist(board_path)), providers=providers)
    color = ort.InferenceSession(str(_must_exist(color_path)), providers=providers)

    # Get the model inputs
    board_inputs: Sequence[ort.NodeArg] = board.get_inputs()
    color_inputs: Sequence[ort.NodeArg] = color.get_inputs()

    # Get the model precision
    board_half = board_inputs[0].type == "tensor(float16)"
    color_half = color_inputs[0].type == "tensor(float16)"

    # Store the shape of the input for later use
    board_input_shape: tuple = board_inputs[0].shape
    board_input_height: int = board_input_shape[2]
    board_input_width: int = board_input_shape[3]
    color_input_shape: tuple = color_inputs[0].shape
    color_input_height: int = color_input_shape[2]
    color_input_width: int = color_input_shape[3]

    models = Models(board=board, color=color, board_inputs=board_inputs, color_inputs=color_inputs)

    return models, (board_input_height, board_input_width), (color_input_height, color_input_width), board_half, color_half



from typing import Optional, cast
from dataclasses import dataclass

import chess
import chess.svg
import chess.pgn


@dataclass
class _Mismatch:
    """
    Small dataclass for mismatched board states between frames.
    """
    location: chess.Square
    color: Optional[bool]
    old_color: Optional[bool]
    capture: bool
    type: Optional[chess.Piece]
    def __iter__(self):
        return iter((self.location, self.color, self.old_color, self.capture, self.type))


class StartChessGame:
    """
    Contains methods and variables to translate model to chess.pgn.Game.
    """
    def __init__(self, board_delay: int = 20) -> None:
        """
        Initializes game model, raw inputs, and stacks for in place mutation.
        """
        self.game: chess.pgn.Game = chess.pgn.Game.without_tag_roster()
        self.chessboard: chess.Board = chess.Board()
        self.old_chessboard: chess.Board = chess.Board()
        self.node: chess.pgn.GameNode = self.game
        self.delay: int = board_delay
        self.move_stack: list[list[chess.Move]] = [[] for _ in range(board_delay)]
        self.board_stack: list[dict[chess.Square, chess.Color]] = [dict() for _ in range(board_delay)]

    def board_has_changed(self) -> bool:
        """
        Outputs equality of new raw map and chessboard map.
        """
        return  self.board_stack[-1].keys() != self.chessboard.piece_map().keys()

    def update_move_stack(self) -> None:
        """
        Removes first index of move stack and adds empty list to top of stack.
        """
        self.move_stack.pop(0)
        self.move_stack.append([])

    def update_board_stack(self, raw_map) -> None:
        """
        Removes first index of board stack and adds latest chessboard to top of stack.
        """
        self.board_stack.pop(0)
        self.board_stack.append(raw_map)

    def find_mismatches(self) -> list[_Mismatch]:
        """
        Returns info of squares whose most recent map mismatch the
        current game state map.
        """
        mismatches: list[_Mismatch] = []
        chessboard_map: dict[chess.Square, chess.Piece] = self.chessboard.piece_map()
        raw_map: dict[chess.Square, chess.Color] = self.board_stack[-1]
        for square in chess.SQUARES:
            if square in raw_map and square in chessboard_map:
                if raw_map[square] != chessboard_map[square].color:
                    mismatches.append(_Mismatch(square, raw_map[square], chessboard_map[square].color, True, chessboard_map[square]))
            elif square in raw_map and square not in chessboard_map:
                mismatches.append(_Mismatch(square, raw_map[square], None, False, None))
            elif square not in raw_map and square in chessboard_map:
                mismatches.append(_Mismatch(square, None, chessboard_map[square].color, False, chessboard_map[square]))
        return mismatches

    def generate_moves_from_mismatches(self, mismatches: list[_Mismatch]) -> None:
        """
        Pairs mismatched squares into possible moves and adds
        them to the class instance's move_stack.
        """
        for x in mismatches:
            for y in mismatches:
                if x.color is None and y.color is not None:
                    white_pawn = x.type == chess.Piece.from_symbol('P')
                    white_rank = x.location // 8 == 6 and y.location // 8 == 7
                    black_pawn = x.type == chess.Piece.from_symbol('p')
                    black_rank = x.location // 8 == 1 and y.location // 8 == 0
                    if white_pawn and white_rank or black_pawn and black_rank:
                        self.move_stack[-1].append(chess.Move(from_square=x.location, to_square=y.location, promotion=chess.QUEEN))
                    elif (y.capture or x.color == y.old_color) and (y.color == x.old_color or x.capture):
                        self.move_stack[-1].append(chess.Move(from_square=x.location, to_square=y.location))

    def validate_moves_and_push(self) -> None:
        """
        Verifies moves are legal and observed multiple times then
        pushes move to chessboard and game.
        """
        goal, target = int(self.delay / 2), int(self.delay * 0.8 / 2) # magic number
        for move in self.move_stack[-1]:
            if sum([move in moves for moves in self.move_stack[goal:]]) >= target:
                if move in self.chessboard.legal_moves:
                    self.node = self.node.add_variation(move)
                    self.chessboard.push(move)

    def find_and_validate_takebacks_and_push(self) -> None:
        """
        Finds, validates, and pushes takebacks to chessboard and game.
        """
        goal, target = int(self.delay / 2), int(self.delay * 0.8 / 2) # magic number
        for move in self.move_stack[-1]:
            if sum([move in moves for moves in self.move_stack[goal:]]) >= target:
                if self.chessboard.fen() != chess.STARTING_FEN and self.node.parent is not None:
                    tri_source = self.chessboard.peek().uci()[2:4] == move.uci()[:2]
                    tri_target = self.chessboard.peek().uci()[:2] == move.uci()[2:4]
                    if move.uci()[:2] != self.chessboard.peek().uci()[2:4]:
                        target_check = sum([chess.Move.from_uci(move.uci()[:2] + self.chessboard.peek().uci()[2:4]) in moves for moves in self.move_stack])
                    else:
                        target_check = False
                    if tri_source or (tri_target and target_check):
                        self.node = cast(chess.pgn.ChildNode, self.node)
                        self.node.parent.variations.remove(self.node)
                        self.node = self.node.parent
                        self.chessboard.pop()

    def find_out_of_turn_moves(self) -> list[chess.Move]:
        """
        Identifies moves that likely skipped a turn.
        """
        future_moves: list[chess.Move] = []
        for move in self.move_stack[-1]:
            if sum([move in moves for moves in self.move_stack]) >= self.delay * .8:
                if self.chessboard.turn != self.chessboard.color_at(move.from_square):
                    future_moves.append(move)
        return future_moves

    def find_potentially_skipped_moves(self, future_moves) -> list[chess.Move]:
        """
        Finds all legal moves that pass the turn.
        """
        potential_fixes: list[chess.Move] = []
        for future_move in future_moves:
            for move in self.chessboard.legal_moves:
                dummy_board = self.chessboard.copy()
                dummy_board.push(move)
                if future_move in dummy_board.legal_moves:
                    potential_fixes.append(move)
        return potential_fixes

    def validate_potentially_skipped_moves_and_push(self, potential_fixes: list[chess.Move]) -> None:
        """
        Verifies fix matches raw board chess.Board object before pushing
        to class instance's chessboard and game.
        """
        goal, target = int(self.delay / 2), int(self.delay * 0.8 / 2)
        for move in potential_fixes:
            test1 = sum([move.from_square not in raw_board for raw_board in self.board_stack[goal:]])
            test2 = sum([move.to_square in raw_board for raw_board in self.board_stack[goal:]])
            if test1 >= target and test2 >= target:
                self.node = self.node.add_variation(move)
                self.chessboard.push(move)
                break

    def write_pgn(self, out_file) -> None:
        """
        Overwrites PGN to external file.
        """
        with open(out_file, 'w') as pgn_file:
            pgn_file.write(self.game.accept(chess.pgn.StringExporter(headers=True)))



from pathlib import Path
from typing import Iterator, Union

import cv2

from .processing import image_resize


def iterate_frames(source: Union[str, int], image_size: int = 640) -> Iterator[tuple[int, cv2.typing.MatLike]]:
    """
    Takes in video source and desired frame size and yields frames.
    """
    cap = cv2.VideoCapture(source)
    if not cap.isOpened():
        raise FileNotFoundError(f"Could not open source: {source}")
    index = 0
    try:
        while True:
            ok, frame = cap.read()
            if not ok:
                break
            frame = image_resize(frame, image_size)
            yield index, frame
            index += 1
    finally:
        cap.release()

def default_pgn_out_path(video: Union[Path, str]) -> Path:
    """
    Sets and creates default output path.
    """
    if isinstance(video, Path):
        out = video.stem
    else:
        out = "out"
    runs = Path.cwd() / "runs"
    runs.mkdir(parents=True, exist_ok=True)
    return runs / f"{out}.pgn"



import shutil
import urllib.request
from pathlib import Path
from typing import Optional

CACHE_DIR = Path.home() / ".knightvision"

DEFAULT_MODELS = {
    "board":  "board-detector.onnx",
    "pieces": "piece-detector.onnx",
}

def _model_url(filename: str) -> str:
    """
    Most recent repo release url.
    """
    return f"https://github.com/nickbuice/knightvision/releases/latest/download/{filename}"

def download_models(filename: str, dest: Path, timeout: int = 60) -> None:
    """
    Downloads file from url.
    """
    url = _model_url(filename)
    dest.parent.mkdir(parents=True, exist_ok=True)
    with urllib.request.urlopen(url, timeout=timeout) as r, open(dest, "wb") as f:
        shutil.copyfileobj(r, f)

def find_model(name: str) -> Optional[Path]:
    """
    Searches for pre-existing models.
    """
    if name not in DEFAULT_MODELS:
        raise ValueError(f"Unknown model name '{name}'. Known: {list(DEFAULT_MODELS)}")
    filename = DEFAULT_MODELS[name]
    candidates: list[Path] = [Path.cwd() / "models" / filename, CACHE_DIR / "models" / filename]
    for p in candidates:
        if p.exists():
            return p
    return None

def valid_model(model: Optional[Path]) -> bool:
    """
    Verifies existence of model path.
    """
    if model:
        if model.exists():
            return True
    return False



from typing import Any, Optional, cast, List

import numpy as np
import cv2
import chess


def image_resize(image: cv2.typing.MatLike, new_size: int) -> cv2.typing.MatLike:
    """
    Resizes longest image edge to new size while maintaining aspect ratio.
    """
    h, w = image.shape[0], image.shape[1]
    if w > h:
        scale = new_size / w
        height = int((np.ceil(h * scale / 32)) * 32)
        dim = (new_size, height)
    else:
        scale = new_size / h
        width = int((np.ceil(w * scale / 32)) * 32)
        dim = (width, new_size)
    return cv2.resize(image, dim)

def _letterbox(img: np.ndarray, new_shape: tuple[int, int] = (640, 640)) -> tuple[np.ndarray, tuple[int, int]]:
    """
    Resize and reshape images while maintaining aspect ratio by adding padding.
    """
    shape = img.shape[:2]
    r = min(new_shape[0] / shape[0], new_shape[1] / shape[1])
    new_unpad = int(round(shape[1] * r)), int(round(shape[0] * r))
    dw, dh = (new_shape[1] - new_unpad[0]) / 2, (new_shape[0] - new_unpad[1]) / 2  # wh padding
    if shape[::-1] != new_unpad:  # resize
        img = cv2.resize(img, new_unpad, interpolation=cv2.INTER_LINEAR)
    top, bottom = int(round(dh - 0.1)), int(round(dh + 0.1))
    left, right = int(round(dw - 0.1)), int(round(dw + 0.1))
    img = cv2.copyMakeBorder(img, top, bottom, left, right, cv2.BORDER_CONSTANT, value=(114, 114, 114))
    return img, (top, left)

def preprocess(input_img: cv2.typing.MatLike, input_shapes: tuple[int, int], half: bool = False) -> tuple[np.ndarray, tuple[int, int]]:
    """
    Adjusts input image to expected input for model.
    """
    img = input_img.copy()
    img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
    img, pad = _letterbox(img, input_shapes)
    image_data = np.array(img) / 255.0
    image_data = np.transpose(image_data, (2, 0, 1))  # Channel first
    ndtype = np.half if half else np.single
    image_data = np.expand_dims(image_data, axis=0).astype(ndtype)
    return image_data, pad

def color_postprocess(img: cv2.typing.MatLike, output: list[np.ndarray], pad: tuple[int, int], input_shapes: tuple[int, int]) -> list[list[int]]:
    """
    Post-process the chess pieces prediction.
    """
    img_height, img_width = img.shape[:2]
    input_height, input_width = input_shapes
    confidence_thres, iou_thres = 0.4, 0.7
    outputs = np.transpose(np.squeeze(output[0]))
    rows = outputs.shape[0]
    centers = []
    boxes = []
    scores = []
    class_ids = []
    gain = min(input_height / img_height, input_width / img_width)
    outputs[:, 0] -= pad[1]
    outputs[:, 1] -= pad[0]
    for i in range(rows):
        classes_scores = outputs[i][4:]
        max_score = np.amax(classes_scores)
        if max_score >= confidence_thres:
            class_id = np.argmax(classes_scores)
            x, y, w, h = outputs[i][0], outputs[i][1], outputs[i][2], outputs[i][3]
            left = int((x - w / 2) / gain)
            top = int((y - h / 2) / gain)
            width = int(w / gain)
            height = int(h / gain)
            x = int(x / gain)
            y = int(y / gain)
            class_ids.append(class_id)
            scores.append(max_score)
            boxes.append([left, top, width, height])
            centers.append([x, y, width, height])
    indices = cv2.dnn.NMSBoxes(boxes, scores, confidence_thres, iou_thres)
    piece_data = []
    for i in indices:
        center = centers[i]
        class_id = class_ids[i]
        piece_data.append([int(center[0]), int(center[1]), int(center[2]), int(center[3]), int(class_id)])
    return piece_data

def board_postprocess(img: cv2.typing.MatLike, output: list[np.ndarray], pad: tuple[int, int], input_shapes: tuple[int, int]) -> Any:
    """
    Post-process the chessboard prediction.
    """
    x, protos = output[0], output[1]  # Two outputs: predictions and protos
    pad_h, pad_w = pad
    img_height, img_width = img.shape[:2]
    input_width, input_height = input_shapes
    ratio = min(input_height / img_height, input_width / img_width)
    nm = 32
    conf_threshold = .4
    iou_threshold = .7
    x = np.einsum("bcn->bnc", x)
    x = x[np.amax(x[..., 4:-nm], axis=-1) > conf_threshold]
    x = np.c_[x[..., :4], np.amax(x[..., 4:-nm], axis=-1), np.argmax(x[..., 4:-nm], axis=-1), x[..., -nm:]]
    x = x[cv2.dnn.NMSBoxes(x[:, :4], x[:, 4], conf_threshold, iou_threshold)]
    if len(x) > 0:
        x[..., [0, 1]] -= x[..., [2, 3]] / 2
        x[..., [2, 3]] += x[..., [0, 1]]
        x[..., :4] -= [pad_w, pad_h, pad_w, pad_h]
        x[..., :4] /= ratio
        x[..., [0, 2]] = x[:, [0, 2]].clip(0, img.shape[1])
        x[..., [1, 3]] = x[:, [1, 3]].clip(0, img.shape[0])
        masks = process_mask(protos[0], x[:, 6:], x[:, :4], img.shape)
        return masks
    else:
        return []

def crop_mask(masks, boxes):
    """
    Takes a mask and a bounding box, and returns a mask that is cropped to the bounding box.
    """
    n, h, w = masks.shape
    x1, y1, x2, y2 = np.split(boxes[:, :, None], 4, 1)
    r = np.arange(w, dtype=x1.dtype)[None, None, :]
    c = np.arange(h, dtype=x1.dtype)[None, :, None]
    return masks * ((r >= x1) * (r < x2) * (c >= y1) * (c < y2))

def process_mask(protos, masks_in, bboxes, im0_shape):
    """
    Takes the output of the mask head, and applies the mask to the bounding boxes.
    """
    c, mh, mw = protos.shape
    masks = np.matmul(masks_in, protos.reshape((c, -1))).reshape((-1, mh, mw)).transpose(1, 2, 0)  # HWN
    masks = np.ascontiguousarray(masks)
    masks = scale_mask(masks, im0_shape)  # re-scale mask from P3 shape to original input image shape
    masks = np.einsum("HWN -> NHW", masks)  # HWN -> NHW
    masks = crop_mask(masks, bboxes)
    return np.greater(masks, 0.5)

def scale_mask(masks, im0_shape, ratio_pad=None):
    """
    Takes a mask, and resizes it to the original image size
    """
    im1_shape = masks.shape[:2]
    if ratio_pad is None:
        gain = min(im1_shape[0] / im0_shape[0], im1_shape[1] / im0_shape[1])  # gain  = old / new
        pad = (im1_shape[1] - im0_shape[1] * gain) / 2, (im1_shape[0] - im0_shape[0] * gain) / 2  # wh padding
    else:
        pad = ratio_pad[1]
    top, left = int(round(pad[1] - 0.1)), int(round(pad[0] - 0.1))  # y, x
    bottom, right = int(round(im1_shape[0] - pad[1] + 0.1)), int(round(im1_shape[1] - pad[0] + 0.1))
    masks = masks[top:bottom, left:right]
    masks = cv2.resize(masks, (im0_shape[1], im0_shape[0]), interpolation=cv2.INTER_LINEAR)
    if len(masks.shape) == 2:
        masks = masks[:, :, None]
    return masks

def _mark_corners(segmentation: np.ndarray) -> list[list[int]]:
    """
    Estimates corners from segmentation.
    """
    mask = segmentation.astype(np.uint8)
    mask *= 255
    contours, hierarchy = cv2.findContours(mask, 1, 2)
    cnt = max(contours, key=cv2.contourArea)
    hull = cv2.convexHull(cnt)
    approx = cv2.approxPolyDP(hull, epsilon=50, closed=True)  # Epsilon magic number
    corners = cast(List[List[int]], approx.reshape(-1, 2).astype(int).tolist())
    return corners


def _sort_clockwise(raw_pts: list[list[int]]) -> list[list[int]]:
    """
    Sorts list of points clockwise.
    """
    center = np.mean(raw_pts, axis=0)
    angles = np.arctan2([p[1] - center[1] for p in raw_pts], [p[0] - center[0] for p in raw_pts])
    return [pt for _, pt in sorted(zip(angles, raw_pts))]


def find_chessboard_corners(corner_results_data: list[np.ndarray]) -> tuple[Optional[np.ndarray], Optional[list]]:
    """
    Uses predicted corners to create conversion matrix.
    """
    transformation_matrix = None
    corners = None
    for mask in corner_results_data:
        corners = _mark_corners(mask)
        if len(corners) == 4:
            real = np.array(_sort_clockwise(corners), dtype=np.float32)
            ideal = np.array([[0, 0], [400, 0], [400, 400], [0, 400]], dtype=np.float32)
            transformation_matrix = cv2.getPerspectiveTransform(real, ideal)
    return transformation_matrix, corners


def _adjust_for_angle(y: float, height: float) -> int:
    """
    Shifts y value of piece location down by 'magic number' percent of box height.
    """
    return int(y + 0.30 * height) # Magic number


def _map_points(num: int, board_size: int = 400) -> int:
    """
    Adjusts transformed coordinates to 8x8 grid.
    """
    return int(np.floor(num * 8 / board_size))


def orient_board(piece_results_data: list[list[int]], transformation_matrix: Optional[np.ndarray]) -> str:
    """
    Takes piece model predictions and outputs flag for board orientation.
    """
    rotation = "NONE"
    white_x, white_y, black_x, black_y = [], [], [], []
    if transformation_matrix is not None:
        for piece in piece_results_data:
            real_x, real_y = piece[0], piece[1]
            real_y = _adjust_for_angle(real_y, height=abs(piece[3]))
            ideal = np.matmul(transformation_matrix, [real_x, real_y, 1])
            if 0 <= ideal[0] / ideal[2] < 400 and 0 <= ideal[1] / ideal[2] < 400:
                raw_name = int(piece[4])
                ideal_x, ideal_y = _map_points(ideal[1] / ideal[2]), _map_points(ideal[0] / ideal[2])
                if raw_name < 1:
                    black_x.append(ideal_x)
                    black_y.append(ideal_y)
                else:
                    white_x.append(ideal_x)
                    white_y.append(ideal_y)
        average_white_x, average_white_y = np.average(white_x), np.average(white_y)
        average_black_x, average_black_y = np.average(black_x), np.average(black_y)
        if average_white_x > 5 > average_white_y > 2 and average_black_x < 2 < average_black_y < 5:
            rotation =  "ROTATE 90 COUNTERCLOCKWISE"
        elif average_black_x > 5 > average_black_y > 2 and average_white_x < 2 < average_white_y < 5:
            rotation = "ROTATE 90 CLOCKWISE"
        elif average_black_y > 5 > average_black_x > 2 and average_white_y < 2 < average_white_x < 5:
            rotation = "ROTATE 180"
        elif average_white_y > 5 > average_white_x > 2 and average_black_y < 2 < average_black_x < 5:
            rotation = "NONE"
    return rotation


def locate_pieces(piece_results_data: list[list[int]], transformation_matrix: Optional[np.ndarray], rotate_board: str) -> dict[chess.Square, int]:
    """
    Transforms chess piece image data into raw numpy board.
    """
    out_board_map = dict()
    piece_names, mapped_pts = [], []
    if transformation_matrix is not None:
        for piece in piece_results_data:
            real_x, real_y = piece[0], piece[1]
            real_y = _adjust_for_angle(real_y, height=abs(piece[3]))
            ideal = np.matmul(transformation_matrix, [real_x, real_y, 1])
            if 0 <= ideal[0] / ideal[2] < 400 and 0 <= ideal[1] / ideal[2] < 400:
                piece_names.append(piece[4])
                mapped_pts.append((_map_points(ideal[1] / ideal[2]), _map_points(ideal[0] / ideal[2])))
        if rotate_board == "ROTATE 90 CLOCKWISE":
            mapped_pts = [(7 - pt[1], pt[0]) for pt in mapped_pts]
        elif rotate_board == "NONE":
            mapped_pts = [(7 - pt[0], 7 - pt[1]) for pt in mapped_pts]
        elif rotate_board == "ROTATE 90 COUNTERCLOCKWISE":
            mapped_pts = [(pt[1], 7 - pt[0]) for pt in mapped_pts]
        for index, (row, column) in enumerate(mapped_pts):
            out_board_map[chess.square(row, column)] = piece_names[index]
    return out_board_map
